# -*- coding: utf-8 -*-
import glob
import os
import os.path as osp

from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import MACCSkeys
from rdkit.Chem import AllChem

import numpy as np

from proteinshake.datasets import Dataset
from proteinshake.utils.pdbbind import parse_pdbbind_PL_index
from proteinshake.utils import extract_tar, download_url

RDLogger.DisableLog('rdApp.*') # disable warnings

class ProteinLigandInterfaceDataset(Dataset):
    """Proteins bound to small molecules from PDBBind with binding site, ligand and affinity information.



    Parameters
    ----------
    root: str
        Root directory where the dataset should be saved.
    name: str
        The name of the dataset.
    version: str
        PDBBind version to use.



    .. code-block:: python

        >>> from proteinshake.datasets import ProteinLigandInterfaceDataset
        >>> dataset = ProteinLigandInterfaceDataset()
        >>> protein = next(dataset.proteins())
        >>> protein['protein']['ligand_smiles']
        'NC(=O)CC[C@H]([NH2+]C[C@@]1(O)OC[C@@H](O)[C@@H](O)[C@@H]1O)C(=O)O'
        >>> protein['protein']['kd']
        99.0
        >>> pocket = [r for r in protein['residues'] if r['binding_site']]



    """


    def __init__(self, version='2020', **kwargs):
        self.version = version

        super().__init__(**kwargs)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/files/*/*_protein.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return filename[:4]

    def download(self):
        download_url(f'https://pdbbind.oss-cn-hangzhou.aliyuncs.com/download/PDBbind_v{self.version}_refined.tar.gz', f'{self.root}/raw')
        extract_tar(f'{self.root}/raw/PDBbind_v{self.version}_refined.tar.gz', f'{self.root}/raw')
        os.rename(f'{self.root}/raw/refined-set', f'{self.root}/raw/files')

        self.index_data = parse_pdbbind_PL_index(osp.join(self.root,
                                                          "raw",
                                                          "files",
                                                          "index",
                                                          f"INDEX_refined_data.{self.version}")
                                                )

    def add_protein_attributes(self, protein):
        pocket = self.pdb2df(f'{self.root}/raw/files/{protein["protein"]["ID"]}/{protein["protein"]["ID"]}_pocket.pdb')
        ligand = Chem.MolFromMolFile(osp.join(self.root,
                                              'raw',
                                              'files',
                                              protein['protein']['ID'],
                                              f'{protein["protein"]["ID"]}_ligand.sdf')
                                     )

        if ligand is None:
            return None
        smiles = Chem.MolToSmiles(ligand)
        fp_morgan = list(map(int, AllChem.GetMorganFingerprintAsBitVect(ligand, 2, nBits=1024).ToBitString()))
        fp_maccs = list(map(int, MACCSkeys.GenMACCSKeys(ligand).ToBitString()))

        pocket_res = np.array(pocket.loc[pocket['atom_type'] == 'CA']['residue_number'].tolist())
        protein_res = np.array(protein['residue']['residue_number']).reshape(-1, 1)

        pocket_atom = np.array(pocket['residue_number'].tolist())
        protein_atom = np.array(protein['atom']['residue_number']).reshape(-1, 1)

        is_site_res = np.zeros_like(protein_res)
        is_site_atom  = np.zeros_like(protein_atom)

        is_site_res[(pocket_res == protein_res).sum(axis=1).nonzero()] = 1.
        is_site_atom[(pocket_atom == protein_atom).sum(axis=1).nonzero()] = 1.


        protein['residue']['binding_site'] = list(map(int, is_site_res.squeeze()))
        protein['atom']['binding_site'] = list(map(int, is_site_atom.squeeze()))

        bind_data = self.index_data[protein['protein']['ID']]
        protein['protein']['kd'] = bind_data['kd']['value']
        protein['protein']['neglog_aff'] = bind_data['neglog_aff']
        protein['protein']['resolution'] = bind_data['resolution']
        protein['protein']['year'] = bind_data['date']
        protein['protein']['ligand_id'] = bind_data['ligand_id']
        protein['protein']['ligand_smiles'] = smiles

        protein['protein']['fp_maccs'] = fp_maccs
        protein['protein']['fp_morgan_r2'] = fp_morgan

        return protein

    def describe(self):
        desc = super().describe()
        desc['property'] = "Small Mol. Binding Site (residue-level)"
        desc['values'] = 2
        desc['type'] = 'Binary'
        return desc



'''
Tests prediction task classes.
'''

import unittest, tempfile
from proteinshake.tasks import *

class TestTasks(unittest.TestCase):

    def task_check(self, task):
        self.assertIsNotNone(task.train_index)
        self.assertIsNotNone(task.train_targets)

    def test_ec_task(self):
        with tempfile.TemporaryDirectory() as tmp:
            task = EnzymeClassTask(root=tmp)
            self.task_check(task)

    def test_binding(self):
        with tempfile.TemporaryDirectory() as tmp:
            task = BindingSiteDetectionTask(root=tmp)
            self.task_check(task)

    def test_affinity(self):
        with tempfile.TemporaryDirectory() as tmp:
            task = LigandAffinityTask(root=tmp)
            self.task_check(task)

    def test_scop(self):
        with tempfile.TemporaryDirectory() as tmp:
            task = StructuralClassTask(root=tmp)
            self.task_check(task)

    def test_retrieve(self):
        with tempfile.TemporaryDirectory() as tmp:
            task = StructureSimilarityTask(root=tmp)
            self.task_check(task)

    # def test_screen(self):
        # with tempfile.TemporaryDirectory() as tmp:
            # task = VirtualScreenTask(root=tmp)
            # self.task_check(task)

if __name__ == '__main__':
    unittest.main()



import pandas as pd
from biopandas.pdb import PandasPdb
from sklearn.neighbors import KDTree

import numpy as np

def get_interfaces(protein, cutoff=6, resolution='residue'):
    """Obtain interfacing residues within a single structure of polymers. Uses
    KDTree data structure for vector search.

    Parameters
    ----------
    protein: dict
        Parsed protein dictionary.

    Returns
    --------
        `list`: indicator list for each residue with 0 if not in interface and 1 else.
    """

    #3-D KD tree
    df = pd.DataFrame({'x': protein[resolution]['x'],
                       'y': protein[resolution]['y'],
                       'z': protein[resolution]['z'],
                       'chain': protein[resolution]['chain_id'],
                       'residue_index':protein[resolution]['residue_number']
                       })
    resi_df = df.groupby(['residue_index', 'chain']).mean().reset_index()
    resi_coords = np.array([resi_df['x'].tolist(), resi_df['y'].tolist(), resi_df['z'].tolist()]).T
    kdt = KDTree(resi_coords, leaf_size=1)

    query = kdt.query_radius(resi_coords, cutoff)
    interface = set()
    for i,result in enumerate(query):
        res_index = resi_df.iloc[i].name
        this_chain = resi_df.iloc[i].chain
        for r in result:
            that_resi = resi_df.iloc[r].name
            that_chain = resi_df.iloc[r].chain
            if this_chain != that_chain:
                interface.add(res_index)
                interface.add(that_resi)

    resi_interface = []
    for r in protein[resolution]['residue_number']:
        resi_interface.append(r in interface)
    return resi_interface



import re

def affinity_parse(s):
    """ Parse the affinity string. e.g. `Kd=30uM`.
    Parameters
    ----------
    s: str
        Affinity measurement string to parse.
    Returns
    -------
    dict
        Dictionary containing parsed affinity information. `value` key stores
        the float value of the measurement. `operator` is the logical operator
        (e.g. `=`, `>`) applied to the value, `unit` is `uM, nM, pM` and
        `measure` is the type experimental measurement (e.g. `Kd, Ki, IC50`)
    """
    operator = "".join(re.findall(r"[=|<|>|~]", s))
    measures = ['Kd', 'Ki', 'IC50']
    for m in measures:
        if s.startswith(m):
            measure = m
            break
    value = float(re.search(r"\d+[.,]?\d*", s).group())
    unit = re.search(r"[m|u|n|f|p]M", s).group()

    return {'operator': operator,
            'measure': measure,
            'value': value,
            'unit': unit
            }

def parse_pdbbind_PL_index(index_path):
    """
    > INDEX_refined_data.2020
    # ==============================================================================
    # List of the protein-ligand complexes in the PDBbind refined set v.2020
    # 5316 protein-ligand complexes in total, which are ranked by binding data
    # Latest update: July 2021
    # PDB code, resolution, release year, -logKd/Ki, Kd/Ki, reference, ligand name
    # ==============================================================================
    2r58  2.00  2007   2.00  Kd=10mM       // 2r58.pdf (MLY)
    3c2f  2.35  2008   2.00  Kd=10.1mM     // 3c2f.pdf (PRP)
    3g2y  1.31  2009   2.00  Ki=10mM       // 3g2y.pdf (GF4)
    3pce  2.06  1998   2.00  Ki=10mM       // 3pce.pdf (3HP)
    4qsu  1.90  2014   2.00  Kd=10mM       // 4qsu.pdf (TDR)
    4qsv  1.90  2014   2.00  Kd=10mM       // 4qsv.pdf (THM)
    """
    data = {}
    with open(index_path, 'r') as ind_file:
        for line in ind_file:
            if line.startswith("#"):
                continue
            pre, post = line.split("//")
            pdbid, res, date, neglog, kd = pre.split()
            kd = affinity_parse(kd)

            lig_id = post.split("(")[1].rstrip(")")

            # remove peptide ligands
            if lig_id.endswith('-mer'):
                continue
            data[pdbid] = {'resolution': float(res),
                           'date': int(date),
                           'kd': kd,
                           'neglog_aff': float(neglog),
                           'ligand_id': lig_id
                           }

    return data



from scipy.ndimage import rotate
import numpy as np

from proteinshake.transforms import Transform


def _get_coords_array(protein, resolution='residue'):
    """ Get a numpy array of the protein coordinates


    Arguments
    ---------
    protein: dict
        A protein dictionary.

    resolution: str
        The resolution at which to return the coordinates ('residue', or 'atom')

    Returns
    --------
    np.array
        Numpy array of shape (n_(residues or atoms),  3)

    """

    return np.array([protein['residue']['x'],
                     protein['residue']['y'],
                     protein['residue']['z']
                     ]
                   ).T

def _set_coords(protein, coord_array, resoluiton='residue', in_place=True):
    """ Given an Nx3 array of coordinates, set them to the
    protein dictionary coord key.

    Arguments
    ----------
    protein: dict
        Proteinshake protein dict
    resolution: str
        Which resolution to use ('residue' or 'atom')
    coord_array: np.array
        Nx3 numpy coordinate array

    """

    assert len(protein[resolution]['x']) == len(coord_array),\
            "Mismatch of coordinate array shape and protein, check choice of resolution"

    protein[resolution]['x'] = list(coord_array[:,0])
    protein[resolution]['y'] = list(coord_array[:,1])
    protein[resolution]['z'] = list(coord_array[:,2])

class CenterTransform(Transform):
    """ Center the coordinates of a protein at atom and residue level.
    We use the Ca to compute the center for all the atoms.
        """
    def __call__(self, protein):
        coords_res = _get_coords_array(protein, resolution='residue')
        coords_atom = _get_coords_array(protein, resolution='atom')

        center = np.mean(coords_res, axis=0)
        new_coords_res = coords_res - center
        new_coords_atom = coords_atom - center

        _set_coords(protein, new_coords_res, resolution='residue')
        _set_coords(protein, new_coords_atom, resolution='atom')

        return protein


class RandomRotateTransform(Transform):
    """ Apply a random rotation to the coordinate arrays of a protein"""

    def __init__(self, seed=42):
        self.seed = seed
        np.random.seed(self.seed)
        pass

    def __call__(self, protein):
        coords_res = _get_coords_array(protein, resolution='residue')
        coords_atom = _get_coords_array(protein, resolution='atom')

        # choose an angle and axis along which to rotate
        angle = np.random.randint(-180, 180, size=1)[0]
        axes_choice = [(0, 1), (0, 2), (1, 2)]
        axes = axes_choice[np.random.randint(0, len(axes_choices), size=1)[0]]

        coords_res_rot = rotate(coords_res, angle=angle, axes=axes)
        coords_atom_rot = rotate(coords_res, angle=angle, axes=axes)

        _set_coords(protein, coords_res_rot, resolution='residue')
        _set_coords(protein, coord_atom_rot, resolution='atom')

        return protein



# -*- coding: utf-8 -*-
import glob
import os

from biopandas.pdb import PandasPdb
from unittest.mock import patch
from tqdm import tqdm

from proteinshake.datasets.dataset import Dataset, AA_THREE_TO_ONE
from proteinshake.utils import download_url

EXTENDED_AA_THREE_TO_ONE = {
    **AA_THREE_TO_ONE,
    'CYZ': 'C',
    'CYX': 'C',
    'HIP': 'H',
    'HID': 'H',
    'HIE': 'H',
}

class ProteinLigandDecoysDataset(Dataset):
    """ Proteins (targets) from `DUDE-Z <https://pubs.acs.org/doi/10.1021/acs.jcim.0c00598>`_ with a list of decoys and active molecules for each.
    Each molecule is encoded as a SMILES string, meant to be used in a virtual screen setting.
    In this setting a model is given a protein and a ligand and outputs a score reflecting the likelihood
    that the given molecule is a binder. Then, this score is used to sort the union of all the ligands and
    decoys. A good model places true ligands at the top of this list. This is known as enrichment factor
    analysis.
    `data source <https://dudez.docking.org/>`_.


    .. code-block:: python

        >>> from proteinshake.datasets import ProteinLigandDecoysDataset
        >>> dataset = ProteinLigandDecoyDataset()
        >>> protein = next(dataset.proteins())
        >>> protein['ligands_smiles'][:3]
        ['C1=CC(NC2=NC(N3CC[NH2+]CC3)=NC(NC3CCCC3)=N2)=CC=[NH+]1', 'C[C@@H]([NH3+])C1=CC=C(C(=O)NC2=C3C=CNC3=NC=C2)C=C1', 'NC1=NON=C1C1=NC2=CC=CC=C2N1C1=CC=CC=C1']
        >>> protein['ligands_ids'][:3]
        ['CHEMBL575962', 'CHEMBL571948', 'CHEMBL1078426']
        >>> protein['decoys_smiles'][:3]
        ['Cc1nc(C)c(N(C)C)nc1C', 'O=C(c1ccc(Cl)cc1Cl)N1CCC(c2cnc[nH]2)CC1', 'O=C(Nc1ccc(O)cc1)C(Cl)(Cl)Cl']
        >>> protein['decoys_ids'][:3]
        ['ZINC000000002211', 'ZINC000000006000', 'ZINC000000002214']


    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @patch('proteinshake.datasets.dataset.AA_THREE_TO_ONE', EXTENDED_AA_THREE_TO_ONE)
    def pdb2df(self, path):
        return super().pdb2df(path)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/*.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return filename.split(".")[0]

    def download(self):
        targets  = ['AA2AR', 'ABL1', 'ACES', 'ADA', 'ADRB2', 'AMPC', 'ANDR', 'CSF1R', 'CXCR4', 'DEF', 'DRD4', 'EGFR', 'FA7', 'FA10', 'FABP4', 'FGFR1', 'FKB1A', 'GLCM', 'HDAC8', 'HIVPR', 'HMDH', 'HS90A', 'ITAL', 'KITH', 'KIT', 'LCK', 'MAPK2', 'MK01', 'MT1', 'NRAM', 'PARP1', 'PLK1', 'PPARA', 'PTN1', 'PUR2', 'RENI', 'ROCK1', 'SRC', 'THRB', 'TRY1', 'TRYB1', 'UROK', 'XIAP']

        for target_id in tqdm(targets, desc='Downloading'):
            # grab receptor
            download_url(f"https://dudez.docking.org/DOCKING_GRIDS_AND_POSES/{target_id}/rec.crg.pdb", f"{self.root}/raw/", log=False)
            os.rename(f'{self.root}/raw/rec.crg.pdb', f'{self.root}/raw/{target_id}.pdb')
            # grab ligands
            download_url(f"https://dudez.docking.org/property_matched/{target_id}_new_DUDE_1/ligands.smi", f"{self.root}/raw/", log=False)
            os.rename(f'{self.root}/raw/ligands.smi', f'{self.root}/raw/ligands_{target_id}.smi')
            # grab decoys
            download_url(f"https://dudez.docking.org/property_matched/{target_id}_new_DUDE_1/decoys.smi", f"{self.root}/raw/", log=False)
            os.rename(f'{self.root}/raw/decoys.smi', f'{self.root}/raw/decoys_{target_id}.smi')


    def add_protein_attributes(self, protein):
        """ We annotte each protein with a list of decoys and a list of active SMILES strings and molecule IDs.
        """
        target = protein['protein']['ID']
        
        for mode in ['decoys', 'ligands']:
            with open(f"{self.root}/raw/{mode}_{target}.smi", "r") as mols:
                smiles, ids = [], []
                for line in mols:
                    smile, mol_id = line.split()
                    smiles.append(smile)
                    ids.append(mol_id)

            protein['protein'][f'{mode}_smiles'] = smiles
            protein['protein'][f'{mode}_ids'] = ids
            protein['protein'][f'num_{mode}'] = len(ids)

        protein['protein']['num_mols'] = protein['protein']['num_ligands'] + protein['protein']['num_decoys']
        return protein




# -*- coding: utf-8 -*-
import os
import glob

from proteinshake.datasets import Dataset
from proteinshake.utils import get_interfaces, extract_tar, download_url

class ProteinProteinInterfaceDataset(Dataset):
    """Protein-protein complexes from PDBBind with annotated interfaces. Residues
    and atoms in each protein are marked with a boolean `is_interface` to indicate
    residues/atoms defined to belong to the interface of two protein chains.
    The default threshold for determining interface residues is 6 Angstroms.
    See :meth:`proteinshake.utils.get_interfaces` for details.

    Parameters
    ----------
    root: str
        Root directory where the dataset should be saved.
    name: str
        The name of the dataset.
    version: str
        PDBBind database version to use.
    cutoff: float
        Distance in angstroms within which a pair of residues is considered to
        belong to the interface.

    .. code-block:: python

            >>> from proteinshake.datasets import ProteinProteinInterfaceDataset 
            >>> dataset = ProteinProteinInterfaceDataset()
            >>> protein = next(dataset.proteins())
            >>> interface = [r for r in protein['residues'] if r['is_interface']]


    """


    def __init__(self, cutoff=6, version='2020', **kwargs):
        self.version = version
        self.cutoff = cutoff
        super().__init__(**kwargs)

    def get_raw_files(self):
        return glob.glob(f'{self.root}/raw/files/*.pdb')[:self.limit]

    def get_id_from_filename(self, filename):
        return filename[:4]

    def download(self):
        download_url(f'https://pdbbind.oss-cn-hangzhou.aliyuncs.com/download/PDBbind_v{self.version}_PP.tar.gz', f'{self.root}/raw')
        extract_tar(f'{self.root}/raw/PDBbind_v{self.version}_PP.tar.gz', f'{self.root}/raw')
        os.rename(f'{self.root}/raw/PP', f'{self.root}/raw/files')

    def add_protein_attributes(self, protein):
        interface_atoms = get_interfaces(protein, self.cutoff, resolution='atom')
        protein['atom']['is_interface'] = interface_atoms
        protein['residue']['is_interface'] = [val
                                              for val, atom_type in \
                                              zip(interface_atoms, protein['atom']['atom_type']) \
                                              if atom_type == 'CA'
                                              ]
        return protein

    def describe(self):
        desc = super().describe()
        desc['property'] = "Protein-protein interface (residue-level)"
        desc['values'] = 2
        desc['type'] = "Binary"
        return desc



'''
Tests all downloads with 'use_precomputed=False'. The number of downloaded files and the number of parsed files is patched to a small number where possible. However, most datasets require downloading one large file, which takes time. Hence removed from GitHub testing CI (by not naming it according to pytest convention).
'''

import unittest, tempfile
from unittest import mock
from proteinshake.datasets import (RCSBDataset,
                                   GeneOntologyDataset,
                                   EnzymeCommissionDataset,
                                   PfamDataset,
                                   ProteinProteinInterfaceDataset,
                                   ProteinLigandInterfaceDataset,
                                   TMAlignDataset,
                                   AlphaFoldDataset,
                                   SCOPDataset
                                   )
from proteinshake.datasets.alphafold import AF_DATASET_NAMES

class TestDownload(unittest.TestCase):

    @mock.patch('proteinshake.datasets.ProteinLigandInterfaceDataset.limit', return_value=5)
    def test_protein_ligand(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinLigandInterfaceDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.ProteinProteinInterfaceDataset.limit', return_value=5)
    def test_protein_protein(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = ProteinProteinInterfaceDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.RCSBDataset.limit', return_value=5)
    def test_rcsb(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = RCSBDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.GeneOntologyDataset.limit', return_value=5)
    def test_go(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = GeneOntologyDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.EnzymeCommissionDataset.limit', return_value=5)
    def test_go(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = EnzymeCommissionDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.PfamDataset.limit', return_value=5)
    def test_pfam(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = PfamDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.TMAlignDataset.limit', return_value=5)
    def test_pfam(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = TMAlignDataset(root=tmp, use_precomputed=False)

    @mock.patch('proteinshake.datasets.AlphaFoldDataset.limit', return_value=5)
    def test_pfam(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            organism = 'methanocaldococcus jannaschii'
            ds = AlphaFoldDataset(root=tmp, version='v4', organism=organism, use_precomputed=False)

    @mock.patch('proteinshake.datasets.SCOPDataset.limit', return_value=5)
    def test_scop(self, mock):
        with tempfile.TemporaryDirectory() as tmp:
            ds = SCOPDataset(root=tmp, use_precomputed=False)

if __name__ == '__main__':
    unittest.main()



from sklearn import metrics

from proteinshake.datasets import ProteinLigandInterfaceDataset
from proteinshake.tasks import Task

class BindingSiteDetectionTask(Task):
    """ Identify the binding residues of a protein-small molecule binding site. This is a residue-level binary classification task.

    """
    
    DatasetClass = ProteinLigandInterfaceDataset

    @property
    def task_type(self):
        return "binary-classification"

    def target(self, protein):
        return protein['residue']['binding_site']

    def evaluate(self, y_pred):
        return {
                'accuracy': metrics.accuracy_score(self.test_targets, y_pred),
                'mcc': metrics.matthews_corrcoef(self.test_targets, y_pred),
                }
